
/* Chrysalide - Outil d'analyse de fichiers binaires
 * codebuffer.c - équivalent Python du fichier "glibext/gcodebuffer.h"
 *
 * Copyright (C) 2012 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "codebuffer.h"


#include <pygobject.h>


#include <glibext/gcodebuffer.h>


#include "../arch/vmpa.h"



/* Retrouve une ligne au sein d'un tampon avec une adresse. */
static PyObject *py_code_buffer_find_line_by_addr(PyObject *, PyObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un tampon de code.                *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Retrouve une ligne au sein d'un tampon avec une adresse.     *
*                                                                             *
*  Retour      : Instance de la ligne trouvée.                                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_code_buffer_find_line_by_addr(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Trouvailles à retourner     */
    PyObject *py_vmpa;                      /* Localisation version Python */
    int ret;                                /* Bilan de lecture des args.  */
    vmpa2t *addr;                           /* Adresse visée par l'opérat° */
    GCodeBuffer *buffer;                    /* Version native              */
    GBufferLine *line;                      /* Ligne trouvée               */

    ret = PyArg_ParseTuple(args, "O", &py_vmpa);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(py_vmpa, (PyObject *)get_python_vmpa_type());
    if (!ret) return NULL;

    addr = get_internal_vmpa(py_vmpa);
    if (addr == NULL) return NULL;

    buffer = G_CODE_BUFFER(pygobject_get(self));

    line = g_code_buffer_find_line_by_addr(buffer, addr, 0, NULL);
    if (line == NULL) Py_RETURN_NONE;

    result = pygobject_new(G_OBJECT(line));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_code_buffer_type(void)
{
    static PyMethodDef py_code_buffer_methods[] = {
        {
            "find_line_by_addr", (PyCFunction)py_code_buffer_find_line_by_addr,
            METH_VARARGS,
            "find_line_by_addr($self, addr, /)\n--\n\nFind a buffer line with a given address."
        },
        { NULL }
    };

    static PyGetSetDef py_code_buffer_getseters[] = {
        { NULL }
    };

    static PyTypeObject py_code_buffer_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.glibext.CodeBuffer",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT,

        .tp_doc         = "PyChrysalide code buffer",

        .tp_methods     = py_code_buffer_methods,
        .tp_getset      = py_code_buffer_getseters,

    };

    return &py_code_buffer_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.glibext.CodeBuffer'.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_code_buffer(PyObject *module)
{
    PyTypeObject *py_code_buffer_type;      /* Type Python 'CodeBuffer'    */
    int ret;                                /* Bilan d'un appel            */
    PyObject *dict;                         /* Dictionnaire du module      */

    py_code_buffer_type = get_python_code_buffer_type();

    py_code_buffer_type->tp_base = &PyGObject_Type;
    py_code_buffer_type->tp_basicsize = py_code_buffer_type->tp_base->tp_basicsize;

    if (PyType_Ready(py_code_buffer_type) != 0)
        return false;

    Py_INCREF(py_code_buffer_type);
    ret = PyModule_AddObject(module, "CodeBuffer", (PyObject *)py_code_buffer_type);
    if (ret != 0) return false;

    dict = PyModule_GetDict(module);
    pygobject_register_class(dict, "CodeBuffer", G_TYPE_CODE_BUFFER, py_code_buffer_type,
                             Py_BuildValue("(O)", py_code_buffer_type->tp_base));

    return true;

}
