
/* Chrysalide - Outil d'analyse de fichiers binaires
 * post.c - traitements complémentaires à la phase de désassemblage
 *
 * Copyright (C) 2016 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "post.h"


#include <assert.h>


#include "processor.h"
#include "target.h"



/******************************************************************************
*                                                                             *
*  Paramètres  : instr   = instruction ARMv7 à traiter.                       *
*                proc    = représentation de l'architecture utilisée.         *
*                context = contexte associé à la phase de désassemblage.      *
*                format  = accès aux données du binaire d'origine.            *
*                index   = indice de l'opérande précisant le saut.            *
*                type    = type du nouveau simple à mettre en place.          *
*                                                                             *
*  Description : Associe un symbole à la valeur ciblée par un opérande.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void post_process_target_resolution(GArchInstruction *instr, GArchProcessor *proc, GProcContext *context, GBinFormat *format, size_t index, SymbolType type)
{
    GArchOperand *op;                       /* Opérande numérique en place */
    virt_t addr;                            /* Adresse visée par le saut   */
    MemoryDataSize ptr_size;                /* Taille de l'espace mémoire  */
    GArchOperand *new;                      /* Instruction de ciblage      */
    vmpa2t target;
    mrange_t trange;                        /* Etendue du symbole à créer  */
    VMPA_BUFFER(loc);
    char name[5 + VMPA_MAX_LEN];
    GBinRoutine *routine;                   /* Nouvelle routine trouvée    */
    GBinSymbol *symbol;                     /* Nouveau symbole construit   */

    op = g_arch_instruction_get_operand(instr, index);
    assert(G_IS_IMM_OPERAND(op));

    if (g_imm_operand_to_virt_t(G_IMM_OPERAND(op), &addr)
        && g_exe_format_translate_address_into_vmpa(G_EXE_FORMAT(format), addr, &target))
    {
        ptr_size = g_arch_processor_get_memory_size(proc);

        new = g_target_operand_new(ptr_size, addr);

        if (!g_target_operand_resolve(G_TARGET_OPERAND(new), format, true))
        {
            vmpa2_virt_to_string(&target, MDS_UNDEFINED, loc, NULL);

            switch (type)
            {
                case STP_ROUTINE:
                    snprintf(name, sizeof(name), "loc_%s", loc + 2);
                    break;

                case STP_CODE_LABEL:
                    snprintf(name, sizeof(name), "sub_%s", loc + 2);
                    break;

                default:
                    assert(false);
                    break;

            }

            routine = g_binary_routine_new();
            g_binary_routine_set_name(routine, strdup(name));

            init_mrange(&trange, &target, 0);
            g_binary_routine_set_range(routine, &trange);

            symbol = g_binary_symbol_new(type);
            g_binary_symbol_attach_routine(symbol, routine);
            g_binary_format_add_symbol(G_BIN_FORMAT(format), symbol);

            g_target_operand_resolve(G_TARGET_OPERAND(new), format, true);

        }

        g_arch_instruction_replace_operand(instr, new, op);

    }

}
